@@ -1,19 +1,21 @@
 import re
 import types
+from node.utils import UNSET
 from .base import (
     factory,
     ExtractionError,
     fetch_value
 )
 from .utils import (
-    UNSET,
     attr_value,
+    generic_html5_attrs,
     cssclasses,
     css_managed_props,
     cssid,
     managedprops,
     vocabulary,
 )
+from .tsf import _
 
 
 ###############################################################################
@@ -67,7 +69,8 @@ factory.doc['props']['required'] = \
 """Whether this value is required or not.
 """
 
-factory.defaults['required_message'] = u'Mandatory field was empty'
+factory.defaults['required_message'] = _('required_message',
+                                         default=u'Mandatory field was empty')
 factory.doc['props']['required_message'] = \
 """Message to be shown if required condition was not met.
 """
@@ -115,6 +118,11 @@ factory.doc['props']['title'] = """\
 Optional help text to be rendered in the title attribute.
 """
 
+factory.defaults['data'] = dict()
+factory.doc['props']['data'] = \
+"""Additional data rendered as HTML5 data attributes on DOM Element.
+"""
+
 factory.defaults['display_proxy'] = False
 factory.doc['props']['display_proxy'] = """\
 If 'True' and widget mode 'display', widget value gets rendered as hidden
@@ -183,6 +191,7 @@ def input_attributes_common(widget, data, excludes=list(), value=None):
         'type': attr_value('type', widget, data),
         'value': value,
     }
+    input_attrs.update(generic_html5_attrs(attr_value('data', widget, data)))
     for attr_name in excludes:
         del input_attrs[attr_name]
     return input_attrs
@@ -201,10 +210,11 @@ def input_attributes_full(widget, data, value=None):
 
 @managedprops('type', 'size', 'disabled', 'autofocus', 'placeholder',
               'autocomplete', *css_managed_props)
-def input_generic_renderer(widget, data):
+def input_generic_renderer(widget, data, custom_attrs={}):
     """Generic HTML ``input`` tag render.
     """
     input_attrs = input_attributes_full(widget, data)
+    input_attrs.update(custom_attrs)
     return data.tag('input', **input_attrs)
 
 
@@ -249,6 +259,7 @@ def generic_display_renderer(widget, data, value=None):
         'id': cssid(widget, 'display'),
         'class_': 'display-%s' % attr_value('class', widget, data) or 'generic'
     }
+    attrs.update(generic_html5_attrs(attr_value('data', widget, data)))
     return data.tag('div', content, **attrs)
 
 
@@ -308,6 +319,7 @@ def tag_renderer(widget, data):
         'id': cssid(widget, 'tag'),
         'class_': cssclasses(widget, data),
     }
+    attrs.update(generic_html5_attrs(attr_value('data', widget, data)))
     tag = attr_value('tag', widget, data)
     text = attr_value('text', widget, data)
     return data.tag(tag, text, **attrs)
@@ -432,7 +444,7 @@ def textarea_attributes(widget, data):
     disabled = attr_value('disabled', widget, data) and 'disabled' or None
     readonly = attr_value('readonly', widget, data) and 'readonly' or None
     required = attr_value('required', widget, data) and 'required' or None
-    return {
+    ta_attrs = {
         'autofocus': autofocus,
         'class_': cssclasses(widget, data),
         'cols': attr_value('cols', widget, data),
@@ -446,6 +458,8 @@ def textarea_attributes(widget, data):
         'rows': attr_value('rows', widget, data),
         'wrap': attr_value('wrap', widget, data),
     }
+    ta_attrs.update(generic_html5_attrs(attr_value('data', widget, data)))
+    return ta_attrs
 
 
 textarea_managed_props = ['autofocus', 'cols', 'disabled', 'placeholder',
@@ -454,11 +468,12 @@ textarea_managed_props = ['autofocus', 'cols', 'disabled', 'placeholder',
 
 
 @managedprops(*textarea_managed_props)
-def textarea_renderer(widget, data):
+def textarea_renderer(widget, data, custom_attrs={}):
     """Renders text area.
     """
     tag = data.tag
     area_attrs = textarea_attributes(widget, data)
+    area_attrs.update(custom_attrs)
     value = fetch_value(widget, data)
     if value is None:
         value = ''
@@ -481,6 +496,8 @@ factory.doc['props']['textarea.wrap'] = \
 """Either ``soft``, ``hard``, ``virtual``, ``physical`` or  ``off``.
 """
 
+factory.defaults['textarea.class'] = 'textarea'
+
 factory.defaults['textarea.cols'] = 80
 factory.doc['props']['textarea.cols'] = \
 """Number of characters.
@@ -533,6 +550,7 @@ def lines_display_renderer(widget, data):
         'id': cssid(widget, 'display'),
         'class_': 'display-%s' % attr_value('class', widget, data) or 'generic'
     }
+    attrs.update(generic_html5_attrs(attr_value('data', widget, data)))
     content = u''
     for line in value:
         content += data.tag('li', line)
@@ -594,7 +612,9 @@ def minlength_extractor(widget, data):
     minlength = attr_value('minlength', widget, data, -1)
     if minlength != -1:
         if len(val) < minlength:
-            message = u'Input must have at least %i characters.' % minlength
+            message = _('minlength_extraction_error',
+                        default=u'Input must have at least ${len} characters.',
+                        mapping={'len': minlength})
             raise ExtractionError(message)
     return val
 
@@ -618,7 +638,9 @@ def ascii_extractor(widget, data):
     try:
         str(val)
     except UnicodeEncodeError:
-        raise ExtractionError(u'Input contains illegal characters.')
+        message = _(u'ascii_extractor_error',
+                    default=u'Input contains illegal characters.')
+        raise ExtractionError(message)
     return val
 
 
@@ -745,7 +767,8 @@ factory.doc['props']['password.strength'] = \
 valid.
 """
 
-factory.defaults['weak_password_message'] = u'Password too weak'
+factory.defaults['weak_password_message'] = _('weak_password_message',
+                                              default=u'Password too weak')
 factory.doc['props']['password.weak_password_message'] = \
 """Message shown if password is not strong enough.
 """
@@ -787,7 +810,15 @@ def checkbox_edit_renderer(widget, data):
         input_attrs['checked'] = input_attrs['value'] and 'checked' or None
     if attr_value('format', widget, data) == 'bool':
         input_attrs['value'] = ''
-    checkbox = tag('input', **input_attrs)
+    with_label = attr_value('with_label', widget, data)
+    if with_label:
+        label = tag('label',
+                    '&nbsp;',
+                    for_=cssid(widget, 'input'),
+                    class_='checkbox_label')
+        checkbox = tag('input', **input_attrs) + label
+    else:
+        checkbox = tag('input', **input_attrs)
     input_attrs = {
         'type': 'hidden',
         'value':  'checkboxexists',
@@ -808,6 +839,7 @@ def checkbox_display_renderer(widget, data):
         content = value
     else:
         vocab = dict(vocabulary(attr_value('vocabulary', widget, data, [])))
+        # XXX: value might be 'True' Looks odd.
         content = vocab[bool(value)]
         if data.tag.translate:
             content = data.tag.translate(content)
@@ -851,6 +883,8 @@ factory.doc['props']['checkbox.format'] = """\
 Data-type of the extracted value. One out of ``bool`` or ``string``.
 """
 
+factory.defaults['checkbox.class'] = 'checkbox'
+
 factory.defaults['checkbox.disabled'] = False
 factory.doc['props']['checkbox.disabled'] = """\
 Flag whether checkbox is disabled.
@@ -862,9 +896,9 @@ Set 'checked' attribute explicit. If not given, compute by value.
 """
 
 factory.defaults['checkbox.vocabulary'] = {
-    True: 'yes',
-    False: 'no',
-    UNSET: 'not set',
+    True: _('yes', default=u'Yes'),
+    False: _('no', default=u'No'),
+    UNSET: _('unset', default=u'Unset'), # XXX: never used right now?
 }
 
 factory.doc['props']['checkbox.vocabulary'] = """\
@@ -873,6 +907,24 @@ used for display of the value. Expected keys are ```True```, ```False``` and
 ```UNSET```.
 """
 
+factory.defaults['checkbox.with_label'] = False
+factory.doc['props']['checkbox.with_label'] = """\
+Render empty label tag after visible checkbox in order to make checkbox UI
+customizable via CSS like so::
+
+    input.large_checkbox {
+        display: none;
+    }
+    input.large_checkbox + label {
+        width: 59px;
+        height: 60px;
+        background: url('/checkbox_large.png');
+    }
+    input.large_checkbox:checked + label {
+        background: url('/checkbox_large_selected.png');
+    }
+"""
+
 factory.defaults['checkbox.required_class'] = 'required'
 
 
@@ -925,7 +977,7 @@ def select_exists_marker(widget, data):
               'listing_label_position', 'listing_tag', 'size',
               'label_checkbox_class', 'label_radio_class',
               *css_managed_props)
-def select_edit_renderer(widget, data):
+def select_edit_renderer(widget, data, custom_attrs={}):
     tag = data.tag
     value = fetch_value(widget, data)
     multivalued = attr_value('multivalued', widget, data)
@@ -959,6 +1011,10 @@ def select_edit_renderer(widget, data):
             'autofocus': autofocus,
             'required': required,
         }
+        select_attrs.update(
+            generic_html5_attrs(
+                attr_value('data', widget, data)))
+        select_attrs.update(custom_attrs)
         if disabled is True:
             select_attrs['disabled'] = 'disabled'
         rendered = tag('select', *optiontags, **select_attrs)
@@ -1005,9 +1061,14 @@ def select_edit_renderer(widget, data):
                                                        label_pos, tag)
             tags.append(tag(item_tag, item,
                             **{'id': cssid(widget, tagtype, key)}))
+        wrapper_attrs = {'id': cssid(widget, tagtype, 'wrapper')}
+        wrapper_attrs.update(
+            generic_html5_attrs(
+                attr_value('data', widget, data)))
+        wrapper_attrs.update(custom_attrs)
         taglisting = tag(listing_tag,
                          *tags,
-                         **{'id': cssid(widget, tagtype, 'wrapper')})
+                         **wrapper_attrs)
         return select_exists_marker(widget, data) + taglisting
 
 
@@ -1027,6 +1088,7 @@ def select_display_renderer(widget, data):
         'id': cssid(widget, 'display'),
         'class_': 'display-%s' % attr_value('class', widget, data) or 'generic'
     }
+    attrs.update(generic_html5_attrs(attr_value('data', widget, data)))
     content = u''
     if multivalued and isinstance(value, basestring):
         value = [value]
@@ -1177,19 +1239,28 @@ def convert_bytes(bytes):
 def input_file_display_renderer(widget, data):
     tag = data.tag
     value = data.value
+    attrs = {
+        'class': cssclasses(widget, data),
+    }
+    attrs.update(generic_html5_attrs(attr_value('data', widget, data)))
     if not value:
-        return tag('div', 'No file')
+        no_file_message = _('no_file', default=u'No file')
+        return tag('div', no_file_message, **attrs)
     file = value['file']
     size = convert_bytes(len(file.read()))
     file.seek(0)
-    filename = value.get('filename', 'Unknown')
-    mimetype = value.get('mimetype', 'Unknown')
+    unknown_message = _('unknown', default=u'Unknown')
+    filename_message = _('filename', default=u'Filename: ')
+    mimetype_message = _('mimetype', default=u'Mimetype: ')
+    size_message = _('size', default=u'Size: ')
+    filename = value.get('filename', unknown_message)
+    mimetype = value.get('mimetype', unknown_message)
     return tag('div',
                tag('ul',
-                   tag('li', tag('strong', 'Filename: '), filename),
-                   tag('li', tag('strong', 'Mimetype: '), mimetype),
-                   tag('li', tag('strong', 'Size: '), size)),
-               class_=cssclasses(widget, data))
+                   tag('li', tag('strong', filename_message), filename),
+                   tag('li', tag('strong', mimetype_message), mimetype),
+                   tag('li', tag('strong', size_message), size)),
+               **attrs)
 
 
 @managedprops(*css_managed_props)
@@ -1235,9 +1306,9 @@ Accepted mimetype.
 """
 
 factory.defaults['file.vocabulary'] = [
-    ('keep', u'Keep Existing file'),
-    ('replace', u'Replace existing file'),
-    ('delete', u'Delete existing file'),
+    ('keep', _('file_keep', default=u'Keep Existing file')),
+    ('replace', _('file_replace', default=u'Replace existing file')),
+    ('delete', _('file_delete', default=u'Delete existing file')),
 ]
 
 
@@ -1315,7 +1386,9 @@ EMAIL_RE = u'^[a-zA-Z0-9\._\-]+@[a-zA-Z0-9\._\-]+.[a-zA-Z0-9]{2,6}$'
 def email_extractor(widget, data):
     val = data.extracted
     if not re.match(EMAIL_RE, val is not UNSET and val or ''):
-        raise ExtractionError(u'Input not a valid email address.')
+        message = _('email_address_not_valid',
+                    default=u'Input not a valid email address.')
+        raise ExtractionError(message)
     return val
 
 
@@ -1351,7 +1424,9 @@ URL_RE += u'.?+=&%@!\-\/]))?$'
 def url_extractor(widget, data):
     val = data.extracted
     if not re.match(URL_RE, val is not UNSET and val or ''):
-        raise ExtractionError(u'Input not a valid web address.')
+        message = _('web_address_not_valid',
+                    default=u'Input not a valid web address.')
+        raise ExtractionError(message)
     return val
 
 
@@ -1405,7 +1480,7 @@ factory.defaults['search.class'] = 'search'
 def number_extractor(widget, data):
     val = data.extracted
     if val is UNSET or val == '':
-        return val
+        return UNSET
     datatype = attr_value('datatype', widget, data)
     if datatype == 'integer':
         convert = int
@@ -1414,24 +1489,54 @@ def number_extractor(widget, data):
     else:
         raise ValueError('Output datatype must be integer or float')
     try:
-        val = convert(val)
+        val = convert(val.replace(',', '.'))
     except ValueError:
-        error = u'Input is not a valid number (%s).' % datatype
-        raise ExtractionError(error)
+        message_datatype_float =  _('datatype_float',
+                                    default=u'floating point')
+        message_datatype_integer = _('datatype_integer', default=u'integer')
+        message = _('input_not_valid_number_datatype',
+                    default=u'Input is not a valid ${datatype}.',
+                    mapping={
+                        'datatype': datatype == 'float' and \
+                            message_datatype_float or \
+                            message_datatype_integer
+                    })
+        raise ExtractionError(message)
     min = attr_value('min', widget, data)
     if min and val < min:
-            raise ExtractionError(u'Value has to be at minimum %s.' % min)
+        message = _('input_number_minimum_value',
+                    default=u'Value has to be at minimum ${min}.',
+                    mapping={'min': min})
+        raise ExtractionError(message)
     max = attr_value('max', widget, data)
     if max and val > max:
-        raise ExtractionError(u'Value has to be at maximum %s.' % max)
+        message = _('input_number_maximum_value',
+                    default=u'Value has to be at maximum ${max}.',
+                    mapping={'max': max})
+        raise ExtractionError(message)
     step = attr_value('step', widget, data)
     if step:
         minimum = min or 0
         if (val - minimum) % step:
-            error = u'Value %s has to be in stepping of %s' % (val, step)
             if minimum:
-                error += ' based on a floor value of %s' % minimum
-            raise ExtractionError(error)
+                message = _(
+                    'input_number_step_and_minimum_value',
+                    default=u'Value ${val} has to be in stepping of ${step} '
+                            u'based on a floor value of ${minimum}',
+                    mapping={
+                        'val': val,
+                        'step': step,
+                        'minimum': minimum,
+                    })
+            else:
+                message = _(
+                    'input_number_step_value',
+                    default=u'Value ${val} has to be in stepping of ${step}',
+                    mapping={
+                        'val': val,
+                        'step': step,
+                    })
+            raise ExtractionError(message)
     return val
 
 
